#!/bin/env python

# File: dmxbleinterface.py
# Date: 20240612
# Author: Robert W.B. Linn

# Connects to the ble device with identifier defined below under _IDENTIFIER (default "DMXINTERFACE").
# Waits for write command to sent to the dmxinterface to control an LED DMX light.

# Return CSV string to the connected client with 2 items "status,msg":
# status - 0=Error, 1=Connected, 2=disconnected, 3=Write
# msg - short info
# Examples for connection ok and error:
# 1,Connected
# 0,[ERROR] Peripheral with identifier ABC not found

# Requires Python package simplepyble: install on the RPi with sudo.
# When running from B4J with the B4JBridge ensure to copy this script to the tempjars folder
# If no permission, set permission from a terminal: sudo chmod +x tempjars

# Run Examples
# Run with defaults
# python dmxinterface.py
# Returns: 1,Connected DMXINTERFACE
#
# Run with debug on and wrong BLE device identifier
# python dmxinterface.py -d 1 -n ABC
# Returns: 0,[ERROR] Peripheral with identifier ABC not found

# TODO
# Remove SimpleBLE message [INFO] SimpleBLE: ...Utils.cpp:33 in initialize_winrt: CoGetApartmentType: cotype=-1, qualifier=0, result=800401F0

# Imports
import argparse
import sys
import time
import json
# Cross-platform bindings for Bluetooth Low Energy (BLE) https://pypi.org/project/simplepyble/
import simplepyble

"""
CONSTANTS
"""
# Limit the number of traceback messages
sys.tracebacklimit = 0

# Set the esp32 ble identifier - must match the name case!
_IDENTIFIER = "DMXBLECONTROLLER"

# Define the uuid's
# Service UUID for the 2 characteristics RX and TX
_UART_UUID  = "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"
# Use for writing to the device (ble device receives data from connected client i.e. this script)
_UART_RX    = "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
# Use for reading data (ble device transmits data to connected client i.e. this script)
_UART_TX    = "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"

# Define status indicator returned to the connected ble client
_STATUS_ERROR = 0
_STATUS_CONNECTED = 1
_STATUS_DISCONNECTED = 2
_STATUS_WRITE = 3

# BLE device scan duration in ms
_SCAN_DURATION = 5000

_PRINT_DEBUG = sys.stdout
_PRINT_OUT = sys.stderr

class BLE:
    """
    BLE class for connecting to the ESP32.
    """   
    _DEBUG = False
    
    """
    Init the instance.
    """
    def __init__(self,
                 identifier,
                 uuid_service,
                 uuid_characteristic_read,
                 uuid_characteristic_write,
                 notification_callback,
                 debug=_DEBUG):
        self._debug = debug
        self._connected = False
        self._identifier = identifier
        self._address = None
        self._uuid_service = uuid_service.lower()
        self._uuid_characteristic_read = uuid_characteristic_read.lower()
        self._uuid_characteristic_write = uuid_characteristic_write.lower()
        self._notification_callback = notification_callback

        if self._debug:
            print(f'[Init] Identifier={self._identifier}, Debug={self._debug}',
                  file = sys.stdout)

    """
    Connect to the BLE peripheral with address given in init.
    """
    def connect(self):
        # Check if already connected = do nothing
        if self._connected:
            return
        
        # BLE Get the adapter
        adapters = simplepyble.Adapter.get_adapters()
        if len(adapters) == 0:
            print(f'{_STATUS_ERROR},[ERROR] No BLE adapters found',
                  file = _PRINT_OUT)
            exit(1)

        # Loop over the adapters and pick the first
        if self._debug:
            for i, adapter in enumerate(adapters):
                print(f'{i}: {adapter.identifier()} [{adapter.address()}]',
                      file = _PRINT_DEBUG)

        choice = 0
        adapter = adapters[choice]
        if self._debug:
            print(f'Selected BLE adapter: {adapter.identifier()} [{adapter.address()}]',
                  file = _PRINT_DEBUG)

        if self._debug:
            adapter.set_callback_on_scan_start(
                lambda: print('Scan started',
                              file = _PRINT_DEBUG))
            adapter.set_callback_on_scan_stop(
                lambda: print('Scan complete',
                              file = _PRINT_DEBUG))
            adapter.set_callback_on_scan_found(
                lambda peripheral: print(f'Found device {peripheral.identifier()} [{peripheral.address().lower()}]',
                                         file = _PRINT_DEBUG))

        # Run BLE Scan
        adapter.scan_for(_SCAN_DURATION)
        peripherals = adapter.scan_get_results()

        # Select the peripheral
        choice = -1
        for i, peripheral in enumerate(peripherals):
            if self._debug:
                print(f'{i}: {peripheral.identifier()} [{peripheral.address()}]',
                      file = _PRINT_DEBUG)
            # Check if the peripheral found has dmxinterface device identifier
            if peripheral.identifier() == self._identifier:
                self._address = peripheral.address()
                # Set choice to value > -1
                choice = i

        if choice == -1:
            print(f'{_STATUS_ERROR},[ERROR] Peripheral with identifier {self._identifier} not found',
                  file = _PRINT_OUT)
            exit(1)

        self.peripheral = peripherals[choice]

        # Connect
        if self._debug:
            print(f'Connecting to: {self.peripheral.identifier()} [{self.peripheral.address()}]',
                  file = _PRINT_DEBUG)
        self.peripheral.connect()

        if self._debug:
            print('Successfully connected, listing services...',
                  file = _PRINT_DEBUG)
        services = self.peripheral.services()
        service_characteristic_pair = []
        for service in services:
            for characteristic in service.characteristics():
                service_characteristic_pair.append((service.uuid(), characteristic.uuid()))

        # Query the user to pick a service/characteristic pair
        for i, (service_uuid, characteristic) in enumerate(service_characteristic_pair):
            if self._debug:
                print(f'{i}: {service_uuid} {characteristic}',
                      file = _PRINT_DEBUG)
            if characteristic == self._uuid_characteristic_read:
                choice = i
                if self._debug:
                    print(f'Service/characteristic found: {i}',
                          file = _PRINT_DEBUG)

        if self._debug:
            print(f'Pairing service/characteristic: {self._uuid_service} {self._uuid_characteristic_read}',
                  file = _PRINT_DEBUG)
        self._uuid_service, self._uuid_characteristic_read = service_characteristic_pair[choice]
        time.sleep(2)
            
        # Start notification using the TX characteristics
        # A callback is used to handle the received data
        self.peripheral.notify(self._uuid_service, self._uuid_characteristic_write, self._notification_callback)
        # notification = self.peripheral.notify(self._uuid_service, self._uuid_characteristic_write, lambda data: print(f"Notification: {data}"))
        
        if self._debug:
            print(f'Connected: {self._identifier}, {self._uuid_service}, {self._uuid_characteristic_read}',
                  file = _PRINT_DEBUG)
        print(f'{_STATUS_CONNECTED},Connected {self._identifier}',
              file = _PRINT_OUT)
        self._connected = True

    """
    Write payload to the connected BLE peripheral.
    """
    def write(self, data):
        if self._connected:
            # The write_request requires the payload as a bytes object.
            try:
                self.peripheral.write_request(self._uuid_service, self._uuid_characteristic_read, str.encode(data))
                print(f'{_STATUS_WRITE},{str.encode(data)}',
                      file = _PRINT_OUT)
            except:
                print(f'{_STATUS_ERROR},[ERROR] Disconnected write_request',
                      file = _PRINT_OUT)
        else:
            print(f'{_STATUS_ERROR},[ERROR] Disconnected write_request',
                  file = _PRINT_OUT)

    """
    Disconnect from the BLE peripheral.
    """
    def disconnect(self):
        if self._connected:    
            if self._debug:
                print(f'Disconnecting from {self.peripheral.identifier()}',
                      file = _PRINT_DEBUG)
            self._connected = False
            self.peripheral.disconnect()
            print(f'{_STATUS_DISCONNECTED},Disconnected',
                  file = _PRINT_OUT)

"""
Define the call for the BLE peripheral notifications.
"""
def notify_callback(data):
    msg = data.decode("utf-8")
    print(msg,
          file = sys.stdout)

"""
MAIN
"""    
if __name__ == "__main__":

    # python dmxinterface.py -n DMXINTERFACE
    parser = argparse.ArgumentParser()        
    parser.add_argument("-n", "--name", default=_IDENTIFIER, type=str, help='name Identifier - Identifier for single device only, default DMXINTERFACE.')
    parser.add_argument("-d", "--debug", default=0, type=int, help='debug 0=Off, 1=On - Debug flag.')

    # Parse the arguments and assign to global
    args = parser.parse_args()

    # Init a new instance of BLE device
    ble = BLE(args.name, _UART_UUID, _UART_RX, _UART_TX, notify_callback, args.debug)

    # Connect to ESP32 via BLE 
    ble.connect()
    
    # Event Loop to process user inputs.
    # The user inputs can be issued manually or via application.
    while True:

        # Wait for the input which is a command
        command = input()
        
        # Handle the command        
        if command == "exit":
            # Disconnect and stop this script
            ble.disconnect()
            exit(1)
        elif command == "connect":
            ble.connect()
        elif command == "disconnect":
            ble.disconnect()
        else:
            # Send the command to the ESP32 via BLE
            ble.write(command)
